@@ -23,9 +23,11 @@
 #include <gtk/gtk.h>
 #include <signal.h>
 #include <sys/time.h>
+#include <sys/stat.h>
 #include <time.h>
 #include <string.h>
 #include <assert.h>
+#include <math.h>
 #include "glib-extra.h"
 #include "config.h"
 #include "terminal.h"
@@ -36,6 +38,7 @@
 #include "msg.h"
 #include "typefinder.h"
 #include "compat.h"
+#include "utils.h"
 #include <vector>
 #include <string>
 #include <list>
@@ -69,15 +72,6 @@ struct Tags
   }
 };
 
-static double
-get_time()
-{
-  timeval tv;
-  gettimeofday (&tv, 0);
-
-  return double (tv.tv_sec) + double (tv.tv_usec) * (1.0 / 1000000.0);
-}
-
 static int
 get_columns()
 {
@@ -133,22 +127,48 @@ filename2uri (string& uri)
   return true;
 }
 
+static string
+uri2filename (const string& uri)
+{
+  GError *err = NULL;
+  char *xfilename = g_filename_from_uri (uri.c_str(), NULL, &err);
+
+  if (err != NULL)
+    {
+      g_critical ("Unable to convert from URI: %s", err->message);
+      g_error_free (err);
+    }
+
+  if (xfilename)
+    {
+      string filename = xfilename;
+      g_free (xfilename);
+
+      return filename;
+    }
+  else
+    {
+      return ""; // fail
+    }
+}
+
 struct Options options;
 
 struct Player : public KeyHandler
 {
   vector<string> uris;
 
-  GstElement    *playbin;
-  GMainLoop     *loop;
+  GstElement   *playbin;
+  GMainLoop    *loop;
 
-  guint          play_position;
-  int            cols;
-  Tags           tags;
-  GstState       last_state;
+  guint         play_position;
+  int           cols;
+  Tags          tags;
+  GstState      last_state;
+  string        old_tag_str;
 
-  gint           video_size_width;
-  gint           video_size_height;
+  double        playback_rate;
+  double        playback_rate_step;
 
   enum
   {
@@ -158,14 +178,18 @@ struct Player : public KeyHandler
   void
   reset_tags (int which_tags)
   {
-    string old_codec = tags.codec;
-    guint old_bitrate = tags.bitrate;
+    Tags old_tags = tags;
     tags = Tags();
 
     if (which_tags == KEEP_CODEC_TAGS)		/* otherwise: RESET_ALL_TAGS */
       {
-	tags.codec = old_codec;
-	tags.bitrate = old_bitrate;
+	tags.codec    = old_tags.codec;
+	tags.vcodec   = old_tags.vcodec;
+	tags.bitrate  = old_tags.bitrate;
+      }
+    else
+      {
+        old_tag_str = "";
       }
   }
 
@@ -183,6 +207,20 @@ struct Player : public KeyHandler
     uris.push_back (uri);
   }
 
+  void
+  set_subtitle (string uri)
+  {
+    if (!gst_uri_is_valid (uri.c_str()))
+      {
+        if (!g_path_is_absolute (uri.c_str()))
+          uri = g_get_current_dir() + string (G_DIR_SEPARATOR + uri);
+
+        if (!filename2uri (uri))
+          return;
+      }
+    g_object_set (G_OBJECT (playbin), "suburi", uri.c_str(), NULL);
+  }
+
   string
   make_n_char_string (string s, guint n)
   {
@@ -198,6 +236,28 @@ struct Player : public KeyHandler
 	*si = ' ';
     return s;
   }
+  string
+  format_tags()
+  {
+    string tag_str;
+
+    if (tags.title != "" || tags.artist != "")
+      tag_str += make_n_char_string ("Title   : " + tags.title, cols / 2 - 1) + " " +
+                 make_n_char_string ("Artist  : " + tags.artist, cols / 2 - 1) + "\n";
+
+    if (tags.album != "" || tags.genre != "")
+      tag_str += make_n_char_string ("Album   : " + tags.album, cols / 2 - 1) + " " +
+	         make_n_char_string ("Genre   : " + tags.genre, cols / 2 - 1) + "\n";
+
+    if (tags.comment != "" || tags.date != "")
+      tag_str += make_n_char_string ("Comment : " + tags.comment, cols / 2 - 1) + " " +
+	         make_n_char_string ("Date    : " + tags.date, cols / 2 - 1) + "\n";
+
+    if (tags.codec != "" || tags.vcodec != "")
+      tag_str += make_n_char_string ("Codec   : " + tags.codec + " (audio) " + ((tags.vcodec != "") ? tags.vcodec + " (video)":""), cols / 2 - 1) + "\n";
+
+    return tag_str;
+  }
 
   void
   display_tags()
@@ -205,22 +265,15 @@ struct Player : public KeyHandler
     if (tags.timestamp > 0)
       if (get_time() - tags.timestamp > 0.5) /* allows us to wait a bit for more tags */
 	{
-	  overwrite_time_display();
-	  Msg::print ("\n");
-	  if (tags.title != "" || tags.artist != "")
-	    Msg::print ("%s %s\n",	  make_n_char_string ("Title   : " + tags.title, cols / 2 - 1).c_str(),
-	                                  make_n_char_string ("Artist  : " + tags.artist, cols / 2 - 1).c_str());
-	  if (tags.album != "" || tags.genre != "")
-	    Msg::print ("%s %s\n",	  make_n_char_string ("Album   : " + tags.album, cols / 2 - 1).c_str(),
-	                                  make_n_char_string ("Genre   : " + tags.genre, cols / 2 - 1).c_str());
-	  if (tags.comment != "" || tags.date != "")
-	    Msg::print ("%s %s\n",	  make_n_char_string ("Comment : " + tags.comment, cols / 2 - 1).c_str(),
-	                                  make_n_char_string ("Date    : " + tags.date, cols / 2 - 1).c_str());
-	  if (tags.codec != "" || tags.vcodec != "" || tags.bitrate != 0)
-	    Msg::print ("%s %s%.1f kbit/s\n",
-	                                  make_n_char_string ("Codec   : " + tags.codec + " (audio) " + ((tags.vcodec != "") ? tags.vcodec + " (video)":""), cols / 2 - 1).c_str(),
-	                                                     ("Bitrate : "), tags.bitrate / 1000.0);
-	  Msg::print ("\n");
+          /* we only display new tags if any of the tags changed */
+          string tag_str = format_tags();
+          if (tag_str != old_tag_str)
+            {
+	      overwrite_time_display();
+	      Msg::print ("\n%s\n", tag_str.c_str());
+              old_tag_str = tag_str;
+            }
+
 	  reset_tags (KEEP_CODEC_TAGS);
 	}
   }
@@ -248,13 +301,13 @@ struct Player : public KeyHandler
     if (uri.substr (0, 5) == "file:")
       {
         /* convert uri "file:///foo/bar" to path "/foo/bar" */
-        string filename = uri.substr (5);
-        while (filename.substr (0, 2) == "//")
-          filename = filename.substr (1);
-
-        TypeFinder tf (filename);
-        if (tf.type() == "image")
-          return true;
+        string filename = uri2filename (uri);
+        if (filename != "")
+          {
+            TypeFinder tf (filename);
+            if (tf.type() == "image")
+              return true;
+          }
       }
     return false;
   }
@@ -265,13 +318,6 @@ struct Player : public KeyHandler
   {
     reset_tags (RESET_ALL_TAGS);
 
-    /*
-     * if we're playing an audio file, we don't need the GtkInterface, so we hide it here
-     * if we're playing a video file, it will be shown again once GStreamer gets to the
-     * point where it sends a "prepare-xwindow-id" message
-     */
-    gtk_interface.hide();
-
     for (;;)
       {
         if (play_position == uris.size() && options.repeat)
@@ -310,11 +356,16 @@ struct Player : public KeyHandler
 
                 gtk_interface.set_title (get_basename (uri));
 
-                video_size_width = 0;
-                video_size_height = 0;
-
                 gst_element_set_state (playbin, GST_STATE_NULL);
                 g_object_set (G_OBJECT (playbin), "uri", uri.c_str(), NULL);
+                if (!options.subtitle)
+                  {
+                    string suburi = guess_subtitle (uri);
+                    if (!suburi.empty())
+                      g_object_set (G_OBJECT (playbin), "suburi", suburi.c_str(), NULL);
+                    else
+                      g_object_set (G_OBJECT (playbin), "suburi", NULL, NULL);
+                  }
                 gst_element_set_state (playbin, GST_STATE_PLAYING);
 
                 if (options.skip > 0)
@@ -334,6 +385,20 @@ struct Player : public KeyHandler
       }
   }
 
+  string
+  guess_subtitle (string uri)
+  {
+    string suburi = uri;
+    unsigned extpos = suburi.rfind ('.'); // File extension position
+
+    if (extpos < suburi.length() - 1)
+      suburi.replace (extpos+1, suburi.npos, "srt");
+    else
+      suburi = "";
+
+    return suburi;
+  }
+
   void
   seek (gint64 new_pos)
   {
@@ -354,8 +419,24 @@ struct Player : public KeyHandler
     if (new_pos < 0)
       new_pos = 0;
 
-    gst_element_seek (playbin, 1.0, GST_FORMAT_TIME, GST_SEEK_FLAG_FLUSH, GST_SEEK_TYPE_SET,
-                      new_pos, GST_SEEK_TYPE_NONE,  GST_CLOCK_TIME_NONE);
+    // Use *_SET and a position for both start and stop, otherwise quick
+    // changes (forward, backward, forward), can leave stop_pos as the reverse
+    // start_pos and once playing forward reaches that point playback stops.
+    gint64 start_pos;
+    gint64 stop_pos;
+    if (playback_rate >= 0)
+      {
+        start_pos = new_pos;
+        stop_pos = GST_CLOCK_TIME_NONE;
+      }
+    else
+      {
+        // when playing in reverse it is from stop to start
+        start_pos = 0;
+        stop_pos = new_pos;
+      }
+    gst_element_seek (playbin, playback_rate, GST_FORMAT_TIME, GST_SEEK_FLAG_FLUSH,
+                      GST_SEEK_TYPE_SET, start_pos, GST_SEEK_TYPE_SET, stop_pos);
   }
 
   void
@@ -368,11 +449,18 @@ struct Player : public KeyHandler
     seek (new_pos_sec * GST_SECOND);
   }
 
+  void
+  set_playback_rate (double rate)
+  {
+    playback_rate = rate;
+    Msg::update_status ("Playback Rate: %.2fx", playback_rate);
+
+    relative_seek (0);
+  }
+
   void
   toggle_pause()
   {
-//    overwrite_time_display();
-
     if (last_state == GST_STATE_PAUSED) {
       gst_element_set_state (playbin, GST_STATE_PLAYING);
     }
@@ -388,13 +476,18 @@ struct Player : public KeyHandler
     g_object_get (G_OBJECT (playbin), "volume", &cur_volume, NULL);
     cur_volume += volume_change;
 
-    overwrite_time_display();
-    Msg::print ("Volume: %4.1f%% \n", cur_volume*100);
+    Msg::update_status ("Volume: %4.1f%%", cur_volume * 100);
 
     if ((cur_volume >= 0) && (cur_volume <= 10))
       g_object_set (G_OBJECT (playbin), "volume", cur_volume, NULL);
   }
 
+  void
+  set_alpha (double alpha_change)
+  {
+    gtk_interface.set_opacity (alpha_change);
+  }
+
   void
   mute_unmute()
   {
@@ -409,18 +502,28 @@ struct Player : public KeyHandler
     gtk_interface.toggle_fullscreen();
   }
 
+  void
+  toggle_subtitle()
+  {
+    int flags;
+    g_object_get (G_OBJECT (playbin), "flags", &flags, NULL);
+    g_object_set (G_OBJECT (playbin), "flags", flags ^ 0x4, NULL);
+
+    Msg::update_status ("Subtitles: %s", (flags & 0x4) ? "off" : "on");
+  }
+
   void
   normal_size()
   {
-    if (video_size_width > 0 && video_size_height > 0)
-      gtk_interface.resize (video_size_width, video_size_height);
-    gtk_interface.unfullscreen();
+    gtk_interface.normal_size();
   }
 
   void
   quit()
   {
-    overwrite_time_display();
+    // End with a newline to preserve the time so the user knows where they
+    // left off.
+    Msg::print ("\n\n");
 
     gst_element_set_state (playbin, GST_STATE_NULL);
     if (loop)
@@ -433,6 +536,8 @@ struct Player : public KeyHandler
 
   Player() : playbin (0), loop(0), play_position (0)
   {
+    playback_rate = 1.0;
+    playback_rate_step = pow (2, 1.0 / 7); // approximately 10%, but 7 steps will make playback rate double
     cols = get_columns();
   }
 };
@@ -473,15 +578,44 @@ collect_tags (const GstTagList *tag_list,
     }
 }
 
+class IdleResizeWindow
+{
+  int     width, height;
+
+public:
+  IdleResizeWindow (int width, int height) :
+    width (width),
+    height (height)
+  {
+  }
+
+  static gboolean
+  callback (gpointer *data)
+  {
+    IdleResizeWindow *self = (IdleResizeWindow *) data;
+    gtk_interface.resize (self->width, self->height);
+    delete self;
+
+    /* do not call me again */
+    return FALSE;
+  }
+};
+
 static void
 caps_set_cb (GObject *pad, GParamSpec *pspec, class Player* player)
 {
+  // this callback doesn't occur in main thread
+
   if (GstCaps *caps = Compat::pad_get_current_caps (GST_PAD (pad)))
     {
-      if (Compat::video_get_size (GST_PAD (pad), &player->video_size_width, &player->video_size_height))
+      int width, height;
+
+      if (Compat::video_get_size (GST_PAD (pad), &width, &height))
         {
-          // resize window to match video size
-          player->normal_size();
+          // resize window to match video size (must run in main thread, so we use an idle handler)
+
+          g_idle_add ((GSourceFunc) IdleResizeWindow::callback,
+                      new IdleResizeWindow (width, height));
         }
       gst_caps_unref (caps);
     }
@@ -520,6 +654,26 @@ collect_print_elements (GstElement *parent, const list<GstElement*>& elements)
     return element_name;
 }
 
+static GstBusSyncReply
+my_sync_bus_callback (GstBus * bus, GstMessage * message, gpointer data)
+{
+  switch (GST_MESSAGE_TYPE (message)) {
+    case GST_MESSAGE_ELEMENT:
+      {
+        if (Compat::is_video_overlay_prepare_window_handle_message (message) && gtk_interface.init_ok())
+          {
+            Compat::video_overlay_set_window_handle (message, gtk_interface.window_xid_nolock());
+          }
+      }
+      break;
+    default:
+      /* unhandled message */
+      break;
+  }
+
+  return GST_BUS_PASS;
+}
+
 static gboolean
 my_bus_callback (GstBus * bus, GstMessage * message, gpointer data)
 {
@@ -574,16 +728,6 @@ my_bus_callback (GstBus * bus, GstMessage * message, gpointer data)
 	player.last_state = state;
       }
       break;
-    case GST_MESSAGE_ELEMENT:
-      {
-        if (Compat::is_video_overlay_prepare_window_handle_message (message) && gtk_interface.init_ok())
-          {
-            // show gtk window to display video in
-            gtk_interface.show();
-            Compat::video_overlay_set_window_handle (message, GDK_WINDOW_XWINDOW (gtk_interface.window()->window));
-          }
-      }
-      break;
     default:
       /* unhandled message */
       break;
@@ -596,10 +740,17 @@ my_bus_callback (GstBus * bus, GstMessage * message, gpointer data)
       g_object_get (G_OBJECT (player.playbin), "video-sink", &videosink, NULL);
       if (videosink && !options.novideo)
         {
-          // Find an sink element that has "force-aspect-ratio" property & set it
-          // to TRUE:
-          GstIterator *iterator = gst_bin_iterate_sinks (GST_BIN (videosink));
-          Compat::iterator_foreach (iterator, force_aspect_ratio, NULL);
+          if (GST_IS_BIN (videosink))
+            {
+              // Find an sink element that has "force-aspect-ratio" property & set it
+              // to TRUE:
+              GstIterator *iterator = gst_bin_iterate_sinks (GST_BIN (videosink));
+              Compat::iterator_foreach (iterator, force_aspect_ratio, NULL);
+            }
+          else
+            {
+              force_aspect_ratio (videosink, NULL);
+            }
 
           if (GstPad* pad = gst_element_get_static_pad (videosink, "sink"))
             {
@@ -608,12 +759,39 @@ my_bus_callback (GstBus * bus, GstMessage * message, gpointer data)
                   caps_set_cb (G_OBJECT (pad), NULL, &player);
                   gst_caps_unref (caps);
                 }
-
-              g_signal_connect (pad, "notify::caps", G_CALLBACK (caps_set_cb), &player);
+              /* connect videosink / videopad "notify::caps" signal
+               *
+               * if we did this before (while playing another video file), skip
+               * this step in order to avoid handling the same signal more than
+               * once
+               */
+              if (!g_signal_handler_find (pad,
+                       /* mask      */ GSignalMatchType (G_SIGNAL_MATCH_FUNC | G_SIGNAL_MATCH_DATA),
+                       /* signal_id */ 0,
+                       /* detail    */ 0,
+                       /* closure   */ 0,
+                       /* func      */ (void *) caps_set_cb,
+                       /* data      */ &player))
+                {
+                  g_signal_connect (pad, "notify::caps", G_CALLBACK (caps_set_cb), &player);
+                }
               gst_object_unref (GST_OBJECT (pad));
             }
           gst_object_unref (GST_OBJECT (videosink));
         }
+      /* show window if necessary (number of video streams > 0 || visualization) */
+      int n_video = 0;
+      GstElement *vis_plugin = NULL;
+      g_object_get (player.playbin, "n-video", &n_video, NULL);
+      g_object_get (player.playbin, "vis-plugin", &vis_plugin, NULL);
+
+      if (gtk_interface.init_ok())
+        {
+          if (n_video || vis_plugin)
+            gtk_interface.show();
+          else
+            gtk_interface.hide();
+        }
     }
 
   /* remove message from the queue */
@@ -661,10 +839,26 @@ cb_print_position (gpointer *data)
 
       glong pos_min = tv_pos.tv_sec / 60;
       glong len_min = tv_len.tv_sec / 60;
+
+      player.overwrite_time_display();
       Msg::print ("\rTime: %01lu:%02lu:%02lu.%02lu", pos_min / 60, pos_min % 60, tv_pos.tv_sec % 60, tv_pos.tv_usec / 10000);
       if (len > 0)   /* streams (i.e. http) have len == -1 */
 	Msg::print (" of %01lu:%02lu:%02lu.%02lu", len_min / 60, len_min % 60, tv_len.tv_sec % 60, tv_len.tv_usec / 10000);
 
+      string message = Msg::status();
+      if (message != "")
+        {
+          Msg::print (" | %s", message.c_str());
+        }
+      else
+        {
+          /* only print bitrate if no status message needs to be shown, in
+           * order to avoid too long output lines
+           */
+          if (player.tags.bitrate > 0)
+            Msg::print (" | Bitrate: %.1f kbit/sec", player.tags.bitrate / 1000.);
+        }
+
       string status, blanks;
       // Print [MUTED] if sound is muted:
       gboolean mute;
@@ -702,10 +896,23 @@ idle_start_player (gpointer *data)
 }
 
 
-static inline bool
-is_directory (const string& path)
+enum FileInfo { FI_DIR, FI_REG, FI_OTHER, FI_ERROR };
+
+static inline FileInfo
+file_info (const string& path)
 {
-  return g_file_test (path.c_str(), G_FILE_TEST_IS_DIR);
+  struct stat st;
+
+  if (lstat (path.c_str(), &st) == 0)
+    {
+      if (S_ISDIR (st.st_mode))
+        return FI_DIR;
+      else if (S_ISREG (st.st_mode))
+        return FI_REG;
+      else
+        return FI_OTHER;
+    }
+  return FI_ERROR;
 }
 
 static vector<string>
@@ -721,12 +928,14 @@ crawl (const string& path)
       while ((name = g_dir_read_name (dir)))
         {
           char *full_name = g_build_filename (path.c_str(), name, NULL);
-          if (is_directory (full_name))
+          FileInfo finfo = file_info (full_name);
+
+          if (finfo == FI_DIR)
             {
               vector<string> subdir_files = crawl (full_name);
               results.insert (results.end(), subdir_files.begin(), subdir_files.end());
             }
-          else
+          else if (finfo == FI_REG)
             {
               results.push_back (full_name);
             }
@@ -776,6 +985,12 @@ Player::process_input (int key)
       case '-':
         set_volume (-0.1);
         break;
+      case 'A':
+        set_alpha (0.1);
+        break;
+      case 'a':
+        set_alpha (-0.1);
+        break;
       case 'M':
       case 'm':
         mute_unmute();
@@ -784,6 +999,10 @@ Player::process_input (int key)
       case 'f':
         toggle_fullscreen();
         break;
+      case 'S':
+      case 's':
+        toggle_subtitle();
+        break;
       case 'N':
       case 'n':
         play_next();
@@ -791,6 +1010,24 @@ Player::process_input (int key)
       case '1':
         normal_size();
         break;
+      case 'r':
+        set_playback_rate (playback_rate * -1);
+        break;
+      case KEY_HANDLER_BACKSPACE:
+        set_playback_rate (1);
+        break;
+      case '[':
+        set_playback_rate (playback_rate / playback_rate_step);
+        break;
+      case ']':
+        set_playback_rate (playback_rate * playback_rate_step);
+        break;
+      case '{':
+        set_playback_rate (playback_rate / 2);
+        break;
+      case '}':
+        set_playback_rate (playback_rate * 2);
+        break;
       case '?':
         print_keyboard_help();
         break;
@@ -812,6 +1049,12 @@ Player::print_keyboard_help()
   printf ("   m                    -     toggle mute/unmute\n");
   printf ("   f                    -     toggle fullscreen (only for videos)\n");
   printf ("   1                    -     normal video size (only for videos)\n");
+  printf ("   A/a                  -     increase/decrease opacity by 10%% (only for videos)\n");
+  printf ("   s                    -     toggle subtitles  (only for videos)\n");
+  printf ("   r                    -     reverse playback\n");
+  printf ("   [ ]                  -     playback rate 10%% faster/slower\n");
+  printf ("   { }                  -     playback rate 2x faster/slower\n");
+  printf ("   Backspace            -     playback rate 1x\n");
   printf ("   n                    -     play next file\n");
   printf ("   q                    -     quit gst123\n");
   printf ("   ?                    -     this help\n");
@@ -822,7 +1065,7 @@ Player::print_keyboard_help()
 void
 Player::add_uri_or_directory (const string& name)
 {
-  if (is_directory (name))          // => play all files in this dir
+  if (file_info (name) == FI_DIR)          // => play all files in this dir
     {
       vector<string> uris = crawl (name);
       for (vector<string>::const_iterator ui = uris.begin(); ui != uris.end(); ui++)
@@ -840,6 +1083,12 @@ main (gint   argc,
 {
   Player player;
 
+  if (XInitThreads() == 0)
+    {
+      fprintf (stderr, "%s: Failed to initialize Xlib support for concurrent threads (XInitThreads() failed).\n", argv[0]);
+      return -1;
+    }
+
   /* Setup options */
   options.parse (argc, argv);
 
@@ -938,7 +1187,17 @@ main (gint   argc,
           g_object_set (G_OBJECT (player.playbin), "audio-sink", audio_sink, NULL);
         }
     }
-  gst_bus_add_watch (gst_pipeline_get_bus (GST_PIPELINE (player.playbin)), my_bus_callback, &player);
+  if (options.initial_volume >= 0)
+    {
+      g_object_set (G_OBJECT (player.playbin), "volume", options.initial_volume / 100, NULL);
+    }
+  if (options.subtitle)
+    {
+      player.set_subtitle (options.subtitle);
+    }
+
+  Compat::setup_bus_callbacks (GST_PIPELINE (player.playbin), my_sync_bus_callback, my_bus_callback, &player);
+
   g_timeout_add (130, (GSourceFunc) cb_print_position, &player);
   g_idle_add ((GSourceFunc) idle_start_player, &player);
   signal (SIGINT, sigint_handler);
